#include <npsoe_iot_kit.h>
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <Wire.h>
#include <delays.h>
#include "DHT.h"

#define LED 2
#define DHTPIN 2
#define DHTTYPE DHT11
#define trigPin 8
#define echoPin A1

BLEServer *pServer = NULL;
BLECharacteristic * pTxCharacteristic;
BLECharacteristic * pTxCharacteristic1;
BLECharacteristic * pTxCharacteristic2;
BLECharacteristic * pTxCharacteristic3;
BLECharacteristic * pTxCharacteristic4;

DHT dht(DHTPIN, DHTTYPE);
DELAYMS_T d1(2000);

bool deviceConnected = false;
bool oldDeviceConnected = false;
float flowRate1 = 0; // aka the txValue
float flowRate2 = 0;
String fr2;
String fr1;
int Reading;
float heartrate;
int count;
int timestamp;
String fr3;
String fr4;
String fr5;

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID           "6E400001-B5A3-F393-E0A9-E50E24DCCA9E" // UART service UUID
#define CHARACTERISTIC_UUID_RX "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX1 "6E400004-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX2 "6E400005-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX3 "6E400006-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX4 "6E400007-B5A3-F393-E0A9-E50E24DCCA9E"

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string rxValue = pCharacteristic->getValue();

      if (rxValue.length() > 0) {
        Serial.println("*********");
        Serial.print("Received Value: ");
        for (int i = 0; i < rxValue.length(); i++){
          Serial.print(rxValue[i]);
        }
        Serial.println();
        Serial.println("*********");

        if(rxValue.compare("*T#") == 0){
          digitalWrite(LED,!digitalRead(LED));
        }
      }
    }
};


void setup() {
  
  Serial.begin(115200);
  pinMode(LED,OUTPUT);
  timestamp = millis(); //timer for heartrate sensor

  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);

  dht.begin();

  // Create the BLE Device
  BLEDevice::init("UART Service jana");

  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pTxCharacteristic = pService->createCharacteristic(
										CHARACTERISTIC_UUID_TX,
										BLECharacteristic::PROPERTY_NOTIFY
									);
                      
  pTxCharacteristic->addDescriptor(new BLE2902());

  pTxCharacteristic1 = pService->createCharacteristic(
                      CHARACTERISTIC_UUID_TX1,
                      BLECharacteristic::PROPERTY_NOTIFY
                    );

  pTxCharacteristic1->addDescriptor(new BLE2902());

  pTxCharacteristic2 = pService->createCharacteristic(
                      CHARACTERISTIC_UUID_TX2,
                      BLECharacteristic::PROPERTY_NOTIFY
                    );

  pTxCharacteristic2->addDescriptor(new BLE2902());

  pTxCharacteristic3 = pService->createCharacteristic(
                      CHARACTERISTIC_UUID_TX3,
                      BLECharacteristic::PROPERTY_NOTIFY
                    );

  pTxCharacteristic3->addDescriptor(new BLE2902());

  pTxCharacteristic4 = pService->createCharacteristic(
                      CHARACTERISTIC_UUID_TX4,
                      BLECharacteristic::PROPERTY_NOTIFY
                    );

  pTxCharacteristic4->addDescriptor(new BLE2902());

  BLECharacteristic * pRxCharacteristic = pService->createCharacteristic(
											 CHARACTERISTIC_UUID_RX,
											BLECharacteristic::PROPERTY_WRITE
										);

  pRxCharacteristic->setCallbacks(new MyCallbacks());

  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  Serial.println("Waiting a client connection to notify...");
}

void loop() {

    if (deviceConnected) 
    {
        //pTxCharacteristic->setValue(&txValue, 1); 
        Reading =  analogRead(A0);
        flowRate1 = map(Reading, 0, 2045, 40, 210);

        //increases count if a heartbeat is detected
        if (Reading > 920)
        {
          count++;
        }

        float h = dht.readHumidity();
        float t = dht.readTemperature();      

        digitalWrite(trigPin, LOW);
        delayMicroseconds(10);
        // Sets the trigPin on HIGH state for 10 micro seconds
        digitalWrite(trigPin, HIGH);
        delayMicroseconds(10);
        digitalWrite(trigPin, LOW);
        // Reads the echoPin, returns the sound wave travel time in microseconds
        long duration = pulseIn(echoPin, HIGH);
        // Calculating the distance
        int d = duration*0.034/2;

        fr2 = (String) flowRate1;
        fr3 = (String) h;
        fr4 = (String) t;
        fr5 = (String) d;

        pTxCharacteristic->setValue(fr2.c_str());
        
        //pTxCharacteristic->setValue("Hello from ESP32!"); // Can only send up to 20 characters at the same time
        pTxCharacteristic->notify();
        //flowRate1+=0.125; // to simulate the changing flowrate value
        delay(250); // bluetooth stack will go into congestion, if too many packets are sent

        pTxCharacteristic4->setValue(fr5.c_str());
        pTxCharacteristic4->notify();
        delay(500);

        //calculating and sending BPM
        if ( (millis()-timestamp) >= 10000)
        {
          heartrate = count*6;
          fr1 = (String) heartrate;
          pTxCharacteristic1->setValue(fr1.c_str());
          pTxCharacteristic1->notify();
          delay(500);
          timestamp = millis();
          count = 0;
        }
        
  if ( d1.dlyExpiredRestart() )
  {
    //Send humidity data 
    pTxCharacteristic2->setValue(fr3.c_str());
    pTxCharacteristic2->notify();
    delay(100);

    //Send temperature data
    pTxCharacteristic3->setValue(fr4.c_str());
    pTxCharacteristic3->notify();
    delay(100);
  }

	  }

    // disconnecting
    if (!deviceConnected && oldDeviceConnected) {
        delay(500); // give the bluetooth stack the chance to get things ready
        pServer->startAdvertising(); // restart advertising
        Serial.println("start advertising");
        oldDeviceConnected = deviceConnected;
    }
    // connecting
    if (deviceConnected && !oldDeviceConnected) {
		// do stuff here on connecting
        oldDeviceConnected = deviceConnected;
    }
}
